import pickle as pkl
from itertools import islice
import treeswift as ts
from smallperm import PseudoRandomPermutation as PRP
import logging
from typing import Any
from itertools import combinations
from pathlib import Path
from dataclasses import dataclass
from teedeelee import DistanceMatrix
import numpy as np
from random import Random
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def trees_have_same_clades(tree1_str: str, tree2_str: str) -> bool:
    """
    Compare two trees by converting them to TreeSwift trees and checking if they have identical clades.

    Args:
        tree1_str: First tree in Newick format
        tree2_str: Second tree in Newick format

    Returns:
        bool: True if trees have identical clades, False otherwise
    """
    try:
        # Parse trees with TreeSwift
        ts_tree1 = ts.read_tree_newick(tree1_str)
        ts_tree2 = ts.read_tree_newick(tree2_str)

        # Get clades for each tree
        clades1 = set()
        clades2 = set()

        # Helper to get clades from a tree
        def get_clades(tree, clades_set):
            for node in tree.traverse_postorder():
                if not node.is_leaf():
                    # Get leaf labels under this node
                    leaves = frozenset(leaf.label for leaf in node.traverse_leaves())
                    clades_set.add(leaves)

        # Get clades for both trees
        get_clades(ts_tree1, clades1)
        get_clades(ts_tree2, clades2)

        # Compare clade sets
        return clades1 == clades2

    except Exception as e:
        logger.error(f"Error comparing trees: {e}")
        logger.error(f"Tree1: {tree1_str}")
        logger.error(f"Tree2: {tree2_str}")
        return False


def distance_matrices_have_same_clades(dist1: Any, dist2: dict) -> bool:
    taxonset = list(dist2.keys())
    for i, j in combinations(taxonset, 2):
        # dist2 uses double-nested dict, dist1 has an api like __getitem__(self, i, j)
        if dist1[i, j] != dist2[i][j]:
            return False
    return True


def get_test_trees(m=16, k=200, seed=0):
    """Generate a sequence of test trees with specified parameters.

    Args:
        m: Number of taxa to include in the restricted trees
        k: Number of trees to sample
        seed: Random seed for reproducibility

    Returns:
        Iterator of tree newick strings
    """
    # Load and prepare the family data
    with open("assets/processed_family.pkl", "rb") as f:
        family = pkl.load(f)

    trees = family[0]
    names = sorted(trees.taxon_set.names())

    # Select subset of taxa
    n = len(names)
    prp = PRP(n, seed)
    subset_names = [names[i] for i in islice(prp, m)]

    mapper = {name: str(i) for i, name in enumerate(subset_names)}

    # Restrict trees to subset of taxa
    trees_restricted = trees.restriction(subset_names)

    # Select subset of trees
    subindices = list(islice(PRP(len(trees_restricted), seed), k))
    trees_sub = [trees_restricted[i] for i in subindices]

    # Return iterator of newick strings
    return ((tree.remap(mapper).newick(), tree.remap(mapper).get_distance_matrix()) for tree in trees_sub), (trees_restricted.get_species_tree().remap(mapper).newick(), trees_restricted.get_species_tree().remap(mapper).get_distance_matrix())


def get_reference_trees(
    subset_names,
    subindices,
    reference_path="/Users/lbq/Downloads/S101/01/truegenetrees",
):
    """Generate a sequence of reference trees matching the test parameters.

    Args:
        subset_names: List of taxa names to include
        subindices: List of indices to select trees
        reference_path: Path to the reference trees file

    Returns:
        Iterator of tree newick strings
    """
    with open(reference_path, "r") as f:
        reference_trees = [ts.read_tree_newick(line.strip()) for line in f]
    
    with open(Path(reference_path).parent / 's_tree.trees', 'r') as f:
        reference_tree_species = ts.read_tree_newick(f.read())

    reference_trees_restricted = [
        t.extract_tree_with(set(subset_names)) for t in reference_trees
    ]

    mapper = {name: str(i) for i, name in enumerate(subset_names)}

    reference_trees_sub = [reference_trees_restricted[i] for i in subindices]
    reference_tree_restricted = reference_tree_species.extract_tree_with(set(subset_names))
    for tree in reference_trees_sub:
        for node in tree.traverse_postorder():
            if not node.is_root():
                node.edge_length = 1
    for tree in reference_trees_sub:
        tree.rename_nodes(mapper)
    reference_tree_restricted.rename_nodes(mapper)
    for node in reference_tree_restricted.traverse_postorder():
        if not node.is_root():
            node.edge_length = 1
    return (
        (tree.newick(), tree.distance_matrix(leaf_labels=True))
        for tree in reference_trees_sub
    ), (reference_tree_restricted.newick(), reference_tree_restricted.distance_matrix(leaf_labels=True))


@dataclass(frozen=True)
class MSCDataPoint:
    distance_matrices: list[DistanceMatrix]
    species_tree: str

    def distance_matrices_numpy(self) -> np.ndarray:
        return np.array([self.to_numpy(mat) for mat in self.distance_matrices])
    
    @staticmethod
    def to_numpy(mat: DistanceMatrix) -> np.ndarray:
        ts_names = set(mat.taxon_set.names())
        for t in range(mat.ntaxa):
            if mat.get_taxon_name(t) not in ts_names:
                raise ValueError(f"Taxon {t} not in taxon set")
        # Get upper triangular entries (excluding diagonal) in row-major order
        triu_entries = []
        for i in range(mat.ntaxa):
            for j in range(i + 1, mat.ntaxa):
                distance = int(mat[str(i), str(j)])
                # Convert to 8-bit binary representation
                binary = [(distance >> bit) & 1 for bit in range(8)]
                triu_entries.append(binary)
        return np.array(triu_entries, dtype=np.bool_)


class MSCDataset:
    def __init__(
        self,
        pkl_path: str,
        k_min_max: tuple[int, int],
        m: int,
        seed: int,
    ) -> None:
        with open(pkl_path, 'rb') as f:
            self.family = pkl.load(f)
        self.nfamily = len(self.family)
        self.k_min, self.k_max = k_min_max
        self.m = m
        self.seed = seed
    
    def __len__(self) -> int:
        return 2 ** 32
    
    def __getitem__(self, idx: int) -> MSCDataPoint:
        # Initialize random state
        rng = Random((idx ^ self.seed) % 2**32)
        
        # Select which family and get trees
        which_family = idx % self.nfamily
        trees = self.family[which_family]
        
        # Get sorted taxon names and select subset
        sorted_names = sorted(trees.taxon_set.names())
        prp_names = PRP(len(sorted_names), rng.randint(0, 2**32))
        subset_names = [sorted_names[i] for i in islice(prp_names, self.m)]
        
        # Create name mapping
        mapper = {name: str(i) for i, name in enumerate(subset_names)}
        
        # Select random subset of trees
        k = rng.randint(self.k_min, self.k_max)
        prp = PRP(len(trees), rng.randint(0, 2**32))
        subindices = list(islice(prp, k))
        subset_trees = [trees[i] for i in subindices]
        
        # Restrict trees to subset of taxa and remap names
        subset_trees_restricted = [tree.restriction(subset_names).remap(mapper) for tree in subset_trees]
        
        # Process species tree
        species_tree = trees.get_species_tree()
        species_tree_restricted = species_tree.restriction(subset_names).remap(mapper)
        
        # Generate distance matrices and return
        distance_matrices = [tree.get_distance_matrix() for tree in subset_trees_restricted]
        return MSCDataPoint(distance_matrices, species_tree_restricted.newick())


PKL_PATH = "assets/processed_family.pkl"

dataset = MSCDataset(PKL_PATH, (200, 340), 16, 0)

for i in range(10):
    print(dataset[0].distance_matrices_numpy().shape)
    print(dataset[0].species_tree)


# if __name__ == "__main__":
#     # Example usage
#     test_trees, (test_tree_species_str, test_tree_species_dist) = get_test_trees()

#     # Get the parameters needed for reference trees
#     with open("assets/processed_family.pkl", "rb") as f:
#         family = pkl.load(f)
#     trees = family[0]
#     names = sorted(trees.taxon_set.names())
#     n = len(names)
#     m = 16
#     k = 200

#     prp = PRP(n, 0)
#     subset_names = [names[i] for i in islice(prp, m)]
#     subindices = list(islice(PRP(len(trees.restriction(subset_names)), 0), k))

#     reference_trees, (reference_tree_species_str, reference_tree_species_dist) = get_reference_trees(subset_names, subindices)

#     # Track failures and successes
#     failures = []
#     successes = []

#     print("\nChecking trees for consistency...")
#     for idx, ((test_tree, test_dist), (reference_tree, reference_dist)) in enumerate(
#         zip(test_trees, reference_trees), 1
#     ):
#         clade_match = trees_have_same_clades(test_tree, reference_tree)
#         dist_match = distance_matrices_have_same_clades(test_dist, reference_dist)

#         if not (clade_match and dist_match):
#             failures.append(
#                 {
#                     "index": idx,
#                     "clade_match": clade_match,
#                     "dist_match": dist_match,
#                     "test_tree": test_tree,
#                     "reference_tree": reference_tree,
#                 }
#             )
#         else:
#             successes.append(
#                 {"index": idx, "test_tree": test_tree, "reference_tree": reference_tree}
#             )
    
#     clade_match = trees_have_same_clades(test_tree_species_str, reference_tree_species_str)
#     dist_match = distance_matrices_have_same_clades(test_tree_species_dist, reference_tree_species_dist)

#     print("For species tree:")
#     print(f"Clade match: {clade_match}")
#     print(f"Distance matrices match: {dist_match}")

#     # Print results
#     print("\n" + "=" * 80)
#     print(f"Results Summary:")
#     print(f"Total trees checked: {k}")
#     print(f"Failures found: {len(failures)}")
#     print(f"Success rate: {((k - len(failures)) / k) * 100:.2f}%")

#     if failures:
#         print("\n" + "=" * 80)
#         print("FAILURES:")
#         print("=" * 80)
#         for fail in failures:
#             print(f"\nFailure at index {fail['index']}:")
#             print(f"Clades match: {fail['clade_match']}")
#             print(f"Distance matrices match: {fail['dist_match']}")
#             print("\nTest tree:")
#             print(fail["test_tree"])
#             print("\nReference tree:")
#             print(fail["reference_tree"])
#             print("-" * 80)

#     # Show sample of successes (up to 5)
#     if successes:
#         print("\n" + "=" * 80)
#         print("SAMPLE OF SUCCESSES (up to 5):")
#         print("=" * 80)
#         for success in successes[:5]:
#             print(f"\nSuccess at index {success['index']}:")
#             print("\nTest tree:")
#             print(success["test_tree"])
#             print("\nReference tree:")
#             print(success["reference_tree"])
#             print("-" * 80)
